//
// Sound.cc
// 
// C++ interface to soundfiles
// based on snd.c
// dpwe@media.mit.edu 1994nov02
//

#include "Sound.h"
#include <assert.h>
#include <stdio.h>
#include <unistd.h>	/* for SEEK_SET etc under SunOS 4 */

int Sound::reset(char modechr, double sampRate, int chans, Format fmt)
{   /* Recreate the snd field */
    DBGFPRINTF((stderr, "Sound0x%lx::reset('%c',%.1f,%d,%d)\n", this, modechr, sampRate, chans, (int)fmt));
    if(modechr != 'r' && modechr != 'w')  {
	// Called illegally
	return 0;
    } else {
	if(snd)
	    sndFree(snd);
	snd = sndNew(NULL);
	snd->snd->samplingRate = sampRate;
	snd->snd->channels   = chans;
	snd->snd->format     = fmt;
	snd->mode = modechr;
	return 1;
    }
}
    
Sound::Sound(char modechr, double sampRate, int chans, Format fmt)
{   /* Create a new sound object.  Default arguments specified for 
       each parameter in header file (the C++ way). */
    DBGFPRINTF((stderr, "Sound0x%lx::Sound()\n", this));
    scanbuf = NULL;
    scanbuflen = 0;
    snd = NULL;
    if(reset(modechr, sampRate, chans, fmt)==0) {
	// Called illegally
	delete this;
    }
}

Sound::~Sound(void)
{   /* close the soundfile, destroy the sound object */
    DBGFPRINTF((stderr, "Sound0x%lx::~Sound()\n", this));
    if(scanbuf)	{
	delete [] scanbuf;
	scanbuf = NULL;
	scanbuflen = 0;
    }
    close();
}

int SfmtSize(int format)
{   /* return the bytes per sample of a particular data format */
    return (format & 0x0F);	// for now, in lower 4 bits of format word
}

int Sound::open(char *filename)
{   /* open file for input, or create for output.  Returns zero on success. */
    char modeStr[2];

    DBGFPRINTF((stderr, "Sound0x%lx::open('%s')\n", this, filename));
    // Make sure the snd object is ready for opening
    close();

    modeStr[0] = snd->mode;	modeStr[1] = '\0';
    if( (snd=sndOpen(snd,filename,modeStr))==NULL)  {
	// rebuild snd so it's not null next time
	return -1;
    } else {
	return 0;
    }
}

void Sound::close(void)
{   /* close the soundfile (and spawn a new snd obj with the same chrs) */
    DBGFPRINTF((stderr, "Sound0x%lx::close()\n", this));
    if(snd) {
	char modechr = snd->mode;
	double sampRate = snd->snd->samplingRate;
	int chans = snd->snd->channels;
	Format fmt = (Format)snd->snd->format;
	reset(modechr, sampRate, chans, fmt);
    }
}

long Sound::readFrames(void *buf, long frames)
{   /* read sound data from file */
    return sndReadFrames(snd, buf, frames);
}

long Sound::writeFrames(void *buf, long frames)
{   /* write sound data to file */
    return sndWriteFrames(snd, buf, frames);
}

long Sound::frameTell(void)
{   /* frame index of next to read */
    return sndFrameTell(snd);
}

long Sound::frameSeek(long frame)
{   /* move file point to abs pos (in sample frames) */
    return sndFrameSeek(snd, frame, SEEK_SET);
}

int Sound::frameFeof(void)
{   /* return nonzero if at EOF */
    return sndFeof(snd);
}

long Sound::MinMaxFrames(long startframe, long frames, float *buf)
{
    int chans = getChannels();
    Format fmt = getFormat();

    setChannels(1);
    setFormat(Sfmt_FLOAT);
    if(scanbuf == NULL || frames > scanbuflen) {
	if(scanbuf)	delete [] scanbuf;
	scanbuf = new float[frames];
	scanbuflen = frames;
    }
    frameSeek(startframe);
    long got=readFrames((void*)scanbuf, frames);
    long i;
    *scanbuf = 0;
    float *pf = scanbuf;
    double v, min = *scanbuf, max = *scanbuf;
    for(i=0; i<got; ++i) {
	v = *pf++;
	if (v > max)	max = v;
	if (v < min)	min = v;
    }
    *((float*)buf) = min;
    *(((float*)buf)+1) = max;
    setChannels(chans);
    setFormat(fmt);
    return got;
}
