// This may look like C code, but it is really -*- C++ -*-
//  
// Sound.h
// 
// C++ interface to soundfiles
// based on snd.c calls
// dpwe@media.mit.edu 1994nov02
//

#ifndef _SOUND_H
#define _SOUND_H

extern "C" {
#include "snd.h"
};

typedef unsigned char byte;

enum  Format {
    Sfmt_SHORT = SFMT_SHORT,
    Sfmt_LONG  = SFMT_LONG,
    Sfmt_FLOAT = SFMT_FLOAT
};

enum  ChanMode {
    Scmd_MONO  = SCMD_MONO,
    Scmd_CHAN0 = SCMD_CHAN0,
    Scmd_CHAN1 = SCMD_CHAN1,
    Scmd_CHAN2 = SCMD_CHAN2,
    Scmd_CHAN3 = SCMD_CHAN3
};

int SfmtSize(int format);
    /* return the bytes per sample of a particular data format */

class Sound {
private:
// Details of the source or destination of the bits
    SOUND *snd;
// private buffer for MinMaxFrames
    float *scanbuf;
    long scanbuflen;
public:
// Constructors & destructor
    Sound(char modechr = 'r', 
	  double sampRate = 44100.0, int chans = 1, Format fmt = Sfmt_SHORT);
                                        // create a new sound object
    ~Sound(void);
    int reset(char modechr = 'r', double sampRate = 44100.0, 
	      int chans = 1, Format fmt = Sfmt_SHORT);  // re-initialize
// Accessors
    int getChannels(void)		{ return sndGetUchans(snd); };
    int setChannels(int chans)		{ sndSetUchans(snd, chans); 
					  return sndGetUchans(snd); };
    int getChanMode(void)		{ return (enum ChanMode)sndGetChanmode(snd); };
    int setChanMode(ChanMode mode)     	{ sndSetChanmode(snd, (int)mode); 
					  return (enum ChanMode)sndGetChanmode(snd); };
    Format getFormat(void)	    { return (enum Format)sndGetUformat(snd);};
    Format setFormat(Format fmt)    { sndSetUformat(snd, (int)fmt);
				      return (enum Format)sndGetUformat(snd);};
    double getSampleRate(void)		{ return sndGetSrate(snd); };
    double setSampleRate(double srate)	{ sndSetSrate(snd, srate);
					  return sndGetSrate(snd); };
    long getFrames(void)		{ return sndGetFrames(snd); };
    long setFrames(long frms)		{ sndSetFrames(snd, frms);
					  return sndGetFrames(snd); };
            // does setFrames make sense?  Not easily changed.
            // But could be handy for hinting the size of an output 
            // file to be written.
    char *getFilename(void)		{ return snd->path; }
    char getMode(void)			{ return snd->mode; };
    char setMode(char modechr)		{  if( /* flags == 0 && */
					      (modechr == 'r'  
					       || modechr=='w')) 
					       snd->mode=modechr;
					   return snd->mode; };
    int isOpen()                        { return snd->isOpen; };
 
// File operations
    int open(char *filename);			// open file for input, 
                                                // or create for output.
                                                // Returns zero on success.
    void close(void);				// Force a file close.
                                                // must be followed by delete
    long readFrames(void *buf, long frames);	// read sound data from file
    long writeFrames(void *buf, long frames);	// write sound data to file
    long frameTell(void);			// frame index of next to read
    long frameSeek(long frame);			// move file point to abs pos.
    int frameFeof(void);			// return nonzero if at EOF

    long MinMaxFrames(long startframe, long frames, float *buf);
    // scan for the minimum and maximum values over the specified range
    // return always as two floats, for now anyway
};

#endif /* _SOUND_H */
