/***********************************************************************\
*	fft.h								*
*   Fast Fourier Transform C library - header file			*
*   Based on RECrandell's						*
*   dpwe 1992apr30	incorporated all billg's modifications		*
*   dpwe 22jan90							*
*   08apr90 With experimental FFT2torl - converse of FFT2real		*
\***********************************************************************/

/*  
    To call an FFT, one must first assign the complex exponential factors:
    A call such as
	e = AssignBasis(NULL, size)
    will set up the complex *e to be the array of cos, sin pairs corresponding
    to total number of complex data = size.   This call allocates the
    (cos, sin) array memory for you.  If you already have such memory
    allocated, pass the allocated pointer instead of NULL.
 */

typedef struct {
    float re, im;
    } complex;

typedef struct lnode
    {
    struct lnode *next;
    long		 size;
    complex	 *table;
    } LNODE;

int PureReal PARG((complex *, long));
int IsPowerOfTwo PARG((long));
complex *FindTable PARG((long));	/* search our list of existing LUT's */
complex *AssignBasis PARG((complex *, long));
void reverseDig PARG((complex *, long, long));
void FFT2dimensional PARG((complex *, long, long, complex *));
void FFT2torl PARG((complex *, long, long, FLOATARG, complex *));
void ConjScale PARG((complex *, long, FLOATARG));
void FFT2real PARG((complex *, long, long, complex *));
void Reals PARG((complex *, long, long, int, complex *));
void PackedReals PARG((complex *, long, long, int, complex *));
void FFT2 PARG((complex *, long, long, complex *));
void FFT2raw PARG((complex *, long, long, long, complex *));
void FFTarb PARG((complex *, complex *, long, complex *));
void DFT PARG((complex *, complex *, long, complex *));
/* simplified routines per billg */
void fft_packed PARG((float *, long));
void ifft_packed PARG((float *, long));
void fft_real PARG((complex *, long));
void ifft_real PARG((complex *, long));
void fft PARG((complex *, long));
void ifft PARG((complex *, long));
void fft_skip PARG((complex *, long, long));
void ifft_skip PARG((complex *, long, long));
void fft_2d PARG((complex *, long, long));
void ifft_2d PARG((complex *, long, long));
