/*
 * sndfbyptr.c
 *
 * Like sndf.c, but handles each file type dynamically at run time
 * with function pointers.
 *
 * 1997jan29 dpwe@icsi.berkeley.edu
 * $Header: /n/crab/da/dpwe/src/tksound/dpwelib/RCS/sndfbyptr.c,v 1.1 1997/02/04 19:01:52 dpwe Exp dpwe $
 */

#include "dpwelib.h"
#include "sndf.h"

typedef struct _sndf_ftype {
    /* default file extension (including ".") */
    char**	sndfExtnPtr;
    /* function to read a header */
    int (*readHdr)(FILE *file, SFSTRUCT *snd, int infoBmax);
    /* function to read extra (dynamically-sized) header information */
    int (*readXInfo)(FILE *file, SFSTRUCT *snd, int done);
    /* function to write a header */
    int (*writeHdr)(FILE *file, SFSTRUCT *snd, int infoBmax);
    /* returns the length in bytes of the header */
    long (*hdrLen)(FILE *file);
    /* returns the offset of the last sample byte in the file */
    long (*lastTell)(FILE *file);
    /* performs any necessary sample conversion (in-place) */
    void (*fixSamps)(FILE *file, void *ptr, int fmt, long nitems, int rw);
    /* default description */
    int id;		/* one of the SFF_ codes */
    char *name;		/* ascii keyword */
} SNDF_FTYPE;

#define SNDF_FTYPE_DEFINED
#include "sndffname.c"

/* for each flagset, define the DFLT_SNDF type, which the 
   dflt_sndf_type ends up pointing to after they have all been defined, 
   and also re-redefine one of the 'hidden' include-file versions of 
   sndfExtn so that it resolves to the plain version, as used by 
   applications (hairy but works). */
#ifdef AIFF
#define DFLT_SNDF sfft_aif
#define SF_AIF_sndfExtn	sndfExtn
#endif /* AIFF */
#ifdef NeXT
#define DFLT_SNDF sfft_nxt
#define SF_NXT_sndfExtn	sndfExtn
#endif /* NeXT */
#ifdef ICSI_NIST
#define DFLT_SNDF sfft_nst
#define SF_NST_sndfExtn	sndfExtn
#endif /* ICSI_NIST */
#ifdef MS_WAVE
#define DFLT_SNDF sfft_wav
#define SF_WAV_sndfExtn	sndfExtn
#endif /* MS_WAVE */
#ifdef IRCAM
#define DFLT_SNDF sfft_irc
#define SF_IRC_sndfExtn sndfExtn
#endif /* IRCAM */
/* default case is PCM */
#ifndef DFLT_SNDF
#define DFLT_SNDF sfft_pcm
#define SF_PCM_sndfExtn sndfExtn
#endif /* DFLT_SNDF */

#define FNPREFIX SF_IRC_
#include "sndfbyptr.h"
#include "sndfirc.c"
static SNDF_FTYPE sfft_irc = {
    &sndfExtn,
    SF_IRC_readHdr,
    SF_IRC_readXInfo,
    SF_IRC_writeHdr,
    SF_IRC_hdrLen, 
    SF_IRC_lastTell,
    SF_IRC_fixSamps,
    SFF_IRCAM,
    "IRCAM"
};

#undef FNPREFIX
#define FNPREFIX SF_NXT_
#include "sndfbyptr.h"
#include "sndfnxt.c"
static SNDF_FTYPE sfft_nxt = {
    &sndfExtn,
    SF_NXT_readHdr,
    SF_NXT_readXInfo,
    SF_NXT_writeHdr, 
    SF_NXT_hdrLen, 
    SF_NXT_lastTell, 
    SF_NXT_fixSamps,
    SFF_NEXT,
    "NeXT"
};

#undef FNPREFIX
#define FNPREFIX SF_AIF_
#include "sndfbyptr.h"
#include "sndfaif.c"
static SNDF_FTYPE sfft_aif = {
    &sndfExtn,
    SF_AIF_readHdr,
    SF_AIF_readXInfo,
    SF_AIF_writeHdr, 
    SF_AIF_hdrLen, 
    SF_AIF_lastTell, 
    SF_AIF_fixSamps, 
    SFF_AIFF, 
    "AIFF"
};

#undef FNPREFIX
#define FNPREFIX SF_WAV_
#include "sndfbyptr.h"
#include "sndfwav.c"
static SNDF_FTYPE sfft_wav = {
    &sndfExtn,
    SF_WAV_readHdr,
    SF_WAV_readXInfo,
    SF_WAV_writeHdr, 
    SF_WAV_hdrLen, 
    SF_WAV_lastTell, 
    SF_WAV_fixSamps,
    SFF_MSWAVE, 
    "MSWAVE"
};

#undef FNPREFIX
#define FNPREFIX SF_NST_
#include "sndfbyptr.h"
#include "sndfnist.c"
static SNDF_FTYPE sfft_nst = {
    &sndfExtn,
    SF_NST_readHdr,
    SF_NST_readXInfo,
    SF_NST_writeHdr, 
    SF_NST_hdrLen, 
    SF_NST_lastTell, 
    SF_NST_fixSamps,
    SFF_NIST, 
    "NIST"
};

#undef FNPREFIX
#define FNPREFIX SF_PCM_
#include "sndfbyptr.h"
#include "sndfpcm.c"
static SNDF_FTYPE sfft_pcm = {
    &sndfExtn,
    SF_PCM_readHdr,
    SF_PCM_readXInfo,
    SF_PCM_writeHdr, 
    SF_PCM_hdrLen, 
    SF_PCM_lastTell, 
    SF_PCM_fixSamps, 
    SFF_RAW, 
    "PCM"
};

static SNDF_FTYPE* sndf_ftypes[] = {
    &sfft_irc,
    &sfft_nxt, 
    &sfft_aif,
    &sfft_wav,
    &sfft_nst,
    &sfft_pcm
};

static SNDF_FTYPE* dflt_sndf_type = &DFLT_SNDF;
static SNDF_FTYPE* _sndf_type = NULL;
/* char *sndfExtn = *DFLT_SNDF.sndfExtnPtr); */
/* char *sndfExtn = ".NONE"; */
/* set by #define of sndfExtn to particular default SF_*_sndfExtn above */
/* (horribly convoluted but seems to work) */

/* Provide actual hooks for programs that call these directly */
#undef sndfExtn
#undef SFReadHdr
#undef SFRdXInfo
#undef SFWriteHdr
#undef SFHeDaLen
#undef SFLastSampTell
#undef FixupSamples

void SF_Settype(FILE *file)
{   /* setup the global fn ptr for the format registered with this file */
    SNDF_FTYPE *sfft = GetSFFType(file);
    if (sfft != NULL)
	_sndf_type = sfft;
    if (_sndf_type == NULL) {
	_sndf_type = dflt_sndf_type;
    }
    sndfExtn = *(_sndf_type->sndfExtnPtr);
}

int SFReadHdr(FILE *file, SFSTRUCT *snd, int infoBmax)
{
    SF_Settype(file);
    return _sndf_type->readHdr(file, snd, infoBmax);
}

int SFRdXInfo(FILE *file, SFSTRUCT *snd, int done)
{
    SF_Settype(file);
    return _sndf_type->readXInfo(file, snd, done);
}

int SFWriteHdr(FILE *file, SFSTRUCT *snd, int infoBmax)
{
    SF_Settype(file);
    return _sndf_type->writeHdr(file, snd, infoBmax);
}

#undef FNPREFIX
#define FNPREFIX _sndf_type->
#include "sndfbyptr.h"
/* Run inside SFOpen type calls */
int SFIdentify(FILE* file, int* pmode);
static SNDF_FTYPE *SFFindFormat(int id);
#define MAYBE_ID(file)	\
    if(_sndf_type == NULL) {						\
        int _sfft, _mode;						\
        if( (_sfft = SFIdentify(file, &_mode)) != SFF_UNKNOWN) {	\
	    SetSFFType(file, SFFindFormat(_sfft));			\
	    SetBytemode(file, _mode);					\
        } else { /* unrecognized, use default */			\
            SetSFFType(file, dflt_sndf_type);				\
        }								\
        SF_Settype(file);						\
    }

/* for writing - check that we have some kind of ftype defined */
#define CHECK_FT	\
     if(_sndf_type == NULL) {						\
	 _sndf_type = dflt_sndf_type;					\
     }

#undef sndfExtn

#define SNDF_PTRSTYLE
#include "sndf.c"

/* Function to choose soundfile format */
int SFFormatToID(char *name)
{   /* Return the soundfile format ID correspoding to this name; 
       -1 if not recognized (==SNDF_UNKNOWN) */
    int nfmts = sizeof(sndf_ftypes)/sizeof(SNDF_FTYPE*);
    int i;

    for (i = 0; i<nfmts; ++i) {
	if (strcmp(name, sndf_ftypes[i]->name) == 0) {
	    return sndf_ftypes[i]->id;
	}
    }
    return SFF_UNKNOWN;
}

/* Function to choose soundfile format */
int SFSetDefaultFormatByName(char *name)
{   /* Set the default soundfile format for future operations 
       according to the one that matches the extension?
       Return 1 on success, 0 if no such format is known. */
    int nfmts = sizeof(sndf_ftypes)/sizeof(SNDF_FTYPE*);
    int i;

    for (i = 0; i<nfmts; ++i) {
	if (strcmp(name, sndf_ftypes[i]->name) == 0) {
	    /* install the new type */
	    dflt_sndf_type = _sndf_type = sndf_ftypes[i];
	    sndfExtn = *(_sndf_type->sndfExtnPtr);
	    return 1;
	}
    }
    return 0;
}

static SNDF_FTYPE *SFFindFormat(int id)
{   /* Find an SFFMT by its ID; return it */
    int nfmts = sizeof(sndf_ftypes)/sizeof(SNDF_FTYPE*);
    int i;
    SNDF_FTYPE *rslt = NULL;

    for (i = 0; i<nfmts; ++i) {
	if (sndf_ftypes[i]->id == id) {
	    rslt = sndf_ftypes[i];
	    return rslt;
	}
    }
    return NULL;
}

int SFSetDefaultFormat(int id)
{   /* Set the default soundfile format according to its id */
    _sndf_type = SFFindFormat(id);
    if(_sndf_type != NULL) {
	dflt_sndf_type = _sndf_type;
	sndfExtn = *(_sndf_type->sndfExtnPtr);
	return 1;
    }
    return 0;
}

int SFSetFormat(FILE *file, int id)
{   /* Set the default soundfile format according to its id */
    SNDF_FTYPE *my_sndf_type = SFFindFormat(id);
    if(my_sndf_type != NULL) {
	SetSFFType(file, my_sndf_type);
	return 1;
    }
    return 0;
}

int SFGetFormat(FILE *file)
{   /* return the soundfile format associated with the file, 
       or the default format if FILE is NULL or unknown */
    SNDF_FTYPE *ftp = GetSFFType(file);
    if(ftp)	return ftp->id;
    else if(_sndf_type != NULL) {
	return _sndf_type->id;
    } else {
	return dflt_sndf_type->id;
    }
}

void SFListFormats(FILE *stream)
{   /* print a list of all known formats to <stream> */
    int nfmts = sizeof(sndf_ftypes)/sizeof(SNDF_FTYPE*);
    int i;

    for (i = 0; i<nfmts; ++i) {
	fprintf(stream, "%s ", sndf_ftypes[i]->name);
    }
}

char *SFFormatName(int id) {
    /* descriptive name for this format */
    SNDF_FTYPE *sfft = SFFindFormat(id);
    if (sfft == NULL)
	return "<UNKNOWN>";
    else
	return sfft->name;
}

/* from sndcvt.c */
static int SFIdentify(FILE *file, int *pmode)
{   /* Figure out the type of a soundfile.  Use *pmode to rtn byteswap mode */
    INT32	magique;
    INT32 	aiffMgc;
    INT32	riffMgc;
    INT32	nistMgc;
    char	*pc;
    int 	bm;

    /* fake up magic word for AIFFs */
    pc = (char *)&aiffMgc;
    *pc++ = 'F';	*pc++ = 'O';	*pc++ = 'R';	*pc++ = 'M';
    /* fake up magic word for WAVs */
    pc = (char *)&riffMgc;
    *pc++ = 'R';	*pc++ = 'I';	*pc++ = 'F';	*pc++ = 'F';
    /* & for NISTs */
    pc = (char *)&nistMgc;
    *pc++ = 'N';	*pc++ = 'I';	*pc++ = 'S';	*pc++ = 'T';

    if( (fread(&magique, sizeof(INT32), 1, file)) != 1)
	SFDie(SFE_RDERR, "SFIdentify");
    if( (bm = MatchByteMode(magique, NXT_MAGIC)) != -1) { 
	*pmode = bm;	return(SFF_NEXT);
    }
    if( (bm = MatchByteMode(magique, IRC_MAGIC)) != -1) { 
	*pmode = bm;	return(SFF_IRCAM);
    }
    if( (bm = MatchByteMode(magique, aiffMgc)) != -1) { 
	*pmode = bm;	return(SFF_AIFF);
    }
    if( (bm = MatchByteMode(magique, riffMgc)) != -1) { 
	*pmode = bm;	return(SFF_MSWAVE);
    }
    if( (bm = MatchByteMode(magique, nistMgc)) != -1) { 
	*pmode = bm;	return(SFF_NIST);
    }
    *pmode = INORDER;
    return(SFF_UNKNOWN);
}
