/***************************************************************\
*   sndfpcm.c			HEADERLESS PCM			*
*   SNDF 'header' functions to access headerless PCM files	*
*   i.e. file is always assumed to be 44k1 stereo 16bit linear  *
*   THIS FILE IS CONDITIONALLY INCLUDED IN sndf.c		*
*   10jul91 after 24sep90 after12aug90 dpwe			*
\***************************************************************/

char *sndfExtn = ".pcm";		/* file extension, if used */

/* header always reads as below.. */
#define DF_PCM_SRT 	SNDF_DFLT_SRATE
#define DF_PCM_CHN 	SNDF_DFLT_CHANS
#define DF_PCM_FMT 	SNDF_DFLT_FORMAT

#define PCM_ENV "PCMFORMAT"	/* Name of an environment string for format */

#ifdef THINK_C
#include <unix.h>
#else
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <unistd.h>
#endif
#include <fcntl.h>

/* for getenv */
#include <stdlib.h>
/* for strchr */
#undef strchr
#include <string.h>

int SFReadHdr(file, snd, infoBmax)
    FILE *file;
    SFSTRUCT *snd;
    int infoBmax;
/* utility to sample the header of an opened file, at most infoBmax info bytes */
    {
    long    infoBsize, infoXtra, bufXtra;
    long    rdsize;
    size_t  num,red;
    long    ft;
    char    *irc_infp, *sdf_infp;
    int     chans,formt,i;
    float   srate;
    char    *eptr, *e;

    if(infoBmax < SFDFLTBYTS)	infoBmax = SFDFLTBYTS;
    infoBmax = (infoBmax)&-4;	/* round DOWN to x4 (can't exceed given bfr) */
    bufXtra = (long)(infoBmax-SFDFLTBYTS);

    ft = ftell(file);
    if( ft > 0 )			/* don't try seek on stdin */
	{
	if(fseek(file,(long)0,SEEK_SET) != 0 ) /* move to start of file */
	    return(SFerror = SFE_RDERR);
	}
    /* Must be ok.  Fabricate sfheader info from defaults */
    srate = DF_PCM_SRT;
    chans = DF_PCM_CHN;
    formt = DF_PCM_FMT;
    /* Check for environment var to override  */
    if( (eptr = getenv(PCM_ENV)) != NULL) {
	/* Format is "setenv PCMFORMAT R8000C1Fs"
	   for short, 1 channel data at 8kHz.
	   "F" should come last. */
	if((e = strchr(eptr, 'R')) || (e = strchr(eptr, 'r'))) {
	    srate = atoi(++e);
	}
	if((e = strchr(eptr, 'C')) || (e = strchr(eptr, 'c'))) {
	    chans = atoi(++e);
	}
	if((e = strchr(eptr, 'F')) || (e = strchr(eptr, 'f'))) {
	    ++e;
	    if(*e=='s' || *e=='S' || strncmp(e, "16", 2)==0) {
		formt = SFMT_SHORT;
	    } else if (*e=='c' || *e=='C' || *e=='8') {
		formt = SFMT_CHAR;
	    } else if (*e=='f' || *e=='F') {
		formt = SFMT_FLOAT;
	    } else if (*e=='l' || *e=='L' || strncmp(e, "32", 2)==0) {
		formt = SFMT_LONG;
	    } else if (*e=='u' || *e=='U') {
		formt = SFMT_ULAW;
	    } else if (*e=='a' || *e=='A') {
		formt = SFMT_ALAW;
	    } else if (*e=='o' || *e=='O') {
		formt = SFMT_CHAR_OFFS;
	    } else if (*e=='d' || *e=='D') {
		formt = SFMT_DOUBLE;
	    } else {
		fprintf(stderr, "sndfpcm: Don't get format in %s: not s/c/f/l/u/a/o/d\n", eptr);
	    }
	}
    }
    /* rewrite into different data structure */
    snd->magic = SFMAGIC;
    snd->headBsize = sizeof(SFSTRUCT);
    snd->dataBsize = GetFileSize(file)-4/* bizarre non header */;
    snd->format = formt;
    snd->channels = chans;
    snd->samplingRate = srate;
    /* grab the bizzarre pad samples */
    fread(snd->info, sizeof(short), 2, file);

    return(SFerror = SFE_OK);
    }

int SFRdXInfo(file, psnd, done)
    FILE *file;
    SFSTRUCT *psnd;
    int done;
/* fetch any extra info bytes needed in view of read header
    Passes open file handle, existing sound struct
    already read 'done' bytes; rest inferred from structure */
    {
    size_t  num;
    long    subhlen;
    long    file_pos;
    int     rem;

    /* will not occur for PCM files */
    return(SFerror = SFE_OK);
    }

int SFWriteHdr(file, snd, infoBmax)
    FILE *file;
    SFSTRUCT *snd;
    int infoBmax;
/*  Utility to write out header to open file
    infoBmax optionally limits the number of info chars to write.
    infoBmax = 0 => write out all bytes implied by .headBsize field
    infoBmax >= SFDFLTBYS => write out only infoBmax infobytes */
    {
    size_t  num;
    long    hedSiz;
    char    *irc_infp, *sdf_infp;
    int     i;
    long    infoBsize,ft;
    short   yuk[2];

    if(snd->magic != SFMAGIC)
	return(SFerror = SFE_NSF);
    if(infoBmax < SFDFLTBYTS || infoBmax > SFInfoSize(snd) )
	infoBmax = SFInfoSize(snd);
    ft = ftell(file);
    if( ft > 0 )			/* don't try seek on pipe */
	{
	if(fseek(file, (long)0, SEEK_SET) != 0 )   /* move to start of file */
	return(SFerror = SFE_WRERR);
	}
    /* write nothing .. */
#ifdef OLDHACK
    /* actually, if I write nothing, I get a bus error when I try to 
       write 32768*2 shorts as the first thing in the file.  I can 
       fix it with this one sample pad (YUKYUKYUK) */
    yuk[0] = yuk[1] = 0;
    fwrite(yuk, sizeof(short), 2, file);
    /* perhaps I should skip over this on reads etc .. */
#endif /* OLDHACK */
    return(SFerror = SFE_OK);
    }

static long SFHeDaLen(file)
    FILE *file;
    /* return data space used for header in file - see SFCloseWrHdr */
    {
    long hlen = 0;

    /* no header, actually - just two bizarre pad samples */
    return(4L);
    }

static long SFLastSampTell(file)	/* return last valid ftell for smps */
    FILE	*file;
    {
    return (0L);	/* punt on this for here - just use EOF */
    }

static void FixupSamples(file, ptr, format, nitems, flag)
    FILE  *file;
    void  *ptr;
    int   format;
    long  nitems;
    int   flag;		/* READ_FLAG or WRITE_FLAG indicates direction */
    {
    }	/* always machine order */
