//
// floatArray.cc
//
// Definition of class of 2D array of floats
// dpwe@interval.com, dpwe@media.mit.edu
// 1994aug11, 1995may09
//

// #define DEBUG

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "dpwedefs.h"
#include "floatArray.h"

floatArray::floatArray(int rows, int cols)  
{   /* create an empty float array */
    arows = rows; acols = cols;
    data = new float [arows*acols]; 
    float *f = data, *g = data+arows*acols; 
    while(f<g) *f++ = 0.0; 
    DBGFPRINTF((stderr, "floatArray(%d,%d)=%lx\n",arows,acols,data));
}

floatArray::floatArray(const floatArray &v)
{   /* copy constructor */
    arows = v.arows; acols = v.acols;
    data = new float [arows*acols]; 
    float *f = data, *g = data+arows*acols, *h = v.data; 
    while(f<g) *f++ = *h++; 
    DBGFPRINTF((stderr, "floatArrayCopy(%d,%d)=%lx\n",arows,acols,data));
}

floatArray::~floatArray()
{ 
    DBGFPRINTF((stderr, "~floatArray(%d,%d)=%lx\n",arows,acols,data));
    delete [] data;
}

floatArray& floatArray::operator = (const floatArray& v)
{
    DBGFPRINTF((stderr, "floatArray=(%d,%d)=%lx\n",arows,acols,data));
    if (this != &v) {
	delete [] data;
	arows = v.arows; acols = v.acols;
	data = new float [arows*acols]; 
	float *f = data, *g = data+arows*acols, *h = v.data; 
	while(f<g) *f++ = *h++; 
        DBGFPRINTF((stderr, ".. becomes (%d,%d)=%lx\n",arows,acols,data));
    }
    return *this;
}

void floatArray::Copy(const floatArray& v)
{   // copy over all the corresponding values, but don't resize
    int commonrows = min(arows, v.Rows());
    int commoncols = min(acols, v.Cols());
    int i,j;
    float *pf;

    for(i=0; i<commoncols; ++i)  {
	pf = &El(0,i);
	for(j=0; j<commonrows; ++j)  {
	    *pf++ = v.Val(j,i);
	}
    }
}

void floatArray::Swap(floatArray& v)
{   // echange our data area with that of the passed array
    float *tmpd;
    int   tmp;

    tmp = arows; arows = v.arows; v.arows = tmp;
    tmp = acols; acols = v.acols; v.acols = tmp;
    tmpd = data; data = v.data; v.data = tmpd;
}

floatArray floatArray::At(int baserow, int basecol, int rows, int cols)
{   /* extract the specified submatrix.  rows and cols default to -1 
       indicating 'to the end'. */
    assert(baserow >= 0 && baserow < arows);
    assert(basecol >= 0 && basecol < acols);
    if(rows < 0)  rows = arows - baserow;
    if(cols < 0)  cols = acols - basecol;
    assert(baserow+rows <= arows);
    assert(basecol+cols <= acols);

    floatArray result(rows, cols);
    float *f = &result.El(0,0), *g = &El(baserow, basecol), *h;
    int i,j;
    for(i = 0; i<result.acols; ++i)  {
        h = g;
        for(j = 0; j<result.arows; ++j)  {
            *f++ = *h++;
        }
        g += arows;
    }
    return result;
}

int floatArray::GetRow(int row, float *d, int base, int len)  
{   /* len defaults to -1 indicating that whole row should be taken. 
	Return the actual number of elements transferred. */
    assert(row >= 0 && row < arows);
    assert(base >=0 && base < acols);
    len = (len < 0 || (len+base) > acols) ? acols-base : len;
    float *f = &El(row,base);
    int i; for(i=0; i<len; ++i)  { 
    	*d++ = *f; f += arows; 
    }
    return len;
}

int floatArray::SetRow(int row, float *d, int base, int len)  
{   /* len defaults to -1 indicating that whole row should be set. 
	Return the actual number of elements transferred. */
    assert(row >= 0 && row < arows);
    assert(base >=0 && base < acols);
    len = (len < 0 || (len+base) > acols) ? acols-base : len;
    float *f = &El(row,base);
    int i; for(i=0; i<len; ++i)  { 
    	 *f = *d++; f += arows; 
    }
    return len;
}

int floatArray::GetCol(int col, float *d, int base, int len)  
{   /* len defaults to -1 indicating that whole col should be taken. 
	Return the actual number of elements transferred. */
    assert(col >= 0 && col < acols);
    assert(base >=0 && base < arows);
    len = (len < 0 || (len+base) > arows) ? arows-base : len;
    float *f = &El(base,col);
    int i; for(i=0; i<len; ++i)  { 
    	*d++ = *f++; 
    }
    return len;
}

int floatArray::SetCol(int col, float *d, int base, int len)  
{   /* len defaults to -1 indicating that whole col should be taken. 
	Return the actual number of elements transferred. */
    assert(col >= 0 && col < acols);
    assert(base >=0 && base < arows);
    len = (len < 0 || (len+base) > arows) ? arows-base : len;
    float *f = &El(base,col);
    int i; for(i=0; i<len; ++i)  { 
    	*f++ = *d++; 
    }
    return len;
}

int floatArray::Sum(float *rslt)
{   /* sum all the columns of the float array into a single vec of float */
    int r, c;
    double sum;
    
    for(c = 0; c<acols; ++c)  {
        sum = 0;
        for(r = 0; r<arows; ++r)  {
            sum += El(r,c);
        }
        *rslt++ = sum;
    }
    return acols;
}

void floatArray::NormRows(void)
{   /* scale each row of the array so its max value is 1 */
    int r,c;
    double max;
    for(r = 0; r < arows; ++r) {
	max = El(r,0);
	for(c = 1; c<acols; ++c) {
	    if(El(r,c) > max) 	max = El(r,c);
	}
	if(max != 0) {
	    for(c = 0; c<acols; ++c) {
		El(r,c) /= max;
	    }
	}
    }
}

void floatArray::NormCols(void)
{   /* scale each col of the array so its max value is 1 */
    int r,c;
    double max;
    for(c = 0; c < acols; ++c) {
	max = El(0,c);
	for(r = 1; r<arows; ++r) {
	    if(El(r,c) > max) 	max = El(r,c);
	}
	if(max != 0) {
	    for(r = 0; r<arows; ++r) {
		El(r,c) /= max;
	    }
	}
    }
}

float floatArray::Min(void)
{   /* find smallest value in array */
    double min = El(0,0);
    int r,c;
    for(c=0; c<acols; ++c) {
	for(r=0; r<arows; ++r) {
	    if(El(r,c) < min) 	min = El(r,c);
	}
    }
    return min;
}

float floatArray::Max(void)
{   /* find largest value in array */
    double max = El(0,0);
    int r,c;
    for(c=0; c<acols; ++c) {
	for(r=0; r<arows; ++r) {
	    if(El(r,c) > max) 	max = El(r,c);
	}
    }
    return max;
}

void floatArray::Norm(void)
{   /* scale whole array so its max value is 1 */
    double max = Max();
    if(max != 0) {
	Scale(1.0/max);
    }
}

void floatArray::Scale(double s)
{   /* scale whole array by specified constant */
    int r,c;
    for(c = 0; c < acols; ++c) {
	for(r = 0; r<arows; ++r) {
	    El(r,c) = s * El(r,c);
	}
    }
}

void floatArray::Offset(double o)
{   /* offset whole array by specified constant */
    int r,c;
    for(c = 0; c < acols; ++c) {
	for(r = 0; r<arows; ++r) {
	    El(r,c) = El(r,c) + o;
	}
    }
}

void floatArray::Print(char *title)
{   /* show the values on the console */
    if(title == 0)  title = "";
    printf("floatArray %s:(%dx%d)\n", title, arows, acols);
    int i,j;
    for(i=0;i<arows;++i)  {
        for(j=0;j<acols;++j)  {
            printf("%f ", El(i,j));
        }
    printf("\n");
    }
}

void floatArray::Resize(int rows, int cols)
{   // resize the array to the newly specified rows and cols.  
    // Preserve as much of the data as practicable.
    floatArray tmp(rows, cols);
    tmp.Clear();
    tmp.Copy(*this);
    Swap(tmp);
}

void floatArray::Clear(void)
{   // Set all the elements of the array to zero
    int i, n = arows*acols;
    float *pf;

    pf = data;
    for(i=0; i<n; ++i)  {
	*pf++ = 0;
    }
}

void floatArray::Transpose(void)
{   // Transpose the data in this array
    floatArray tmp(acols, arows);
    int i,j;
    float *pf = data;

    for(i = 0; i<acols; ++i)  {
	for(j=0; j<arows; ++j)  {
	    tmp.El(i,j) = *pf++;
	}
    }
    // exchange our data for that of the temporary
    Swap(tmp);
    // our old data is destroyed when the local floatArray tmp is deleted
}

extern "C" {
#include <DATUtils.h>
};

int floatArray::loadDAT(const char *filename, int dataset /*=0*/)
{   /* Load the array from a given dataset of a DAT file.
       Return the number of elements loaded, or -1 if bad read. */
    int ds, ch, rw, cl;
    double sr;
    char fm;

    if(DATInfo((char *)filename, &ds, &ch, &rw, &cl, &sr, &fm)==0)  {
	return -1;
    }
    if(dataset >= ds)  {
	/* request for a dataset that doesn't exist */
	return -1;
    }
    /* cols are the fastest rotating dimension in a DAT file, 
       so make an implicit transposition on reading into the floatArray */
    Resize(cl, rw);
    return DATRead((char *)filename, dataset, 0, -1, -1, (void **)&data, 'f', 
		   &rw, &cl, &sr);
}

int floatArray::saveDAT(const char *filename, int dataset /*=0*/, int trunc /*=1*/)
{   /* Save the array to a given dataset of a DAT file.
       Return the number of elements saved, or -1 if bad match. 
       <trunc>!=0 forces any existing file to be deleted first. */
    /* cols are the fastest rotating dimension in a DAT file, 
       so make an implicit transposition on reading into the floatArray */
    return DATWrite((char *)filename, dataset, 0, -1, -1, data, 'f', 
		    acols, arows, 0, 'f', trunc);
}

#ifdef MAIN
void testFloatArray(void)
{   /* test the float array class */
    floatArray a(4,6);
    float b[32];
    int i,n;
    
    a.Print("empty");
    for(i = 0; i<10; ++i)  {
        b[i] = float(i)/10.0;
    }
    a.SetRow(2, b);
    a.SetCol(4, b);
    a.Print("set row, col");
    
    floatArray c = a.At(1,2,2,3);
    c.Print("submatrix from 1, 2: 2x3");
    
    b[4] = -1;
    n = a.GetRow(2, b, 1, 4);
    for(i = 0; i<(n+1); ++i)  printf("%.1f ", b[i]);
    printf("\n");
    a.SetRow(3, b, 0, 5);
    n = a.GetCol(5, b, 1);
    for(i = 0; i<n; ++i)  printf("%.1f ", b[i]);
    printf("\n");
    c = a.At();
    c.Print("final c");
    c.Transpose();
    c.Print("transposed");

    floatArray d(2,5);
    int j;

    for(i = 0; i<5; ++i)
	for(j = 0; j<2; ++j)
	    d.El(j, i) = (i+j)%10;
    d.Print("orig");
    d.Resize(3, 4);
    d.Print("res1");
    d.Resize(4, 3);
    d.Print("res2");
    d.Resize(2, 2);
    d.Print("res3");
    d.Resize(3, 5);
    d.Print("res4");
}

void main(int argc, char **argv)
{
    testFloatArray();
}
#endif /* MAIN */
