//			-*- C++ -*-
// floatArray.h
// 
// Class is a 2-D array of floats.
// Notionally, they are interleaved samples from a set of channels 
// of far greater length.
//
// dpwe@interval.com 1994aug11, dpwe@media.mit.edu 1995may09
//

#ifndef _FLOATARRAY_H
#define _FLOATARRAY_H

class floatArray {   // array organized as interleaved rows i.e. +1 is next row
private:
  float *data;
  int   arows;
  int   acols;
protected:
  void Swap(floatArray &v);	   // swap the data with a different one
public:
  floatArray(int rows, int cols);  // create an empty float array
  floatArray(const floatArray &v); // copy constructor
  floatArray& operator = (const floatArray& v); // assign to existg floatArray
  void Copy(const floatArray& v);	// copy over values but don't resize
  ~floatArray();
  int Rows() const	{ return arows; }
  int Cols() const	{ return acols; }
  float &El(int row, int col)  { return data[col*arows + row]; }
  float Val(int row, int col) const  { return data[col*arows + row]; }
  int GetRow(int row, float *d, int base = 0, int len = -1);
  int SetRow(int row, float *d, int base = 0, int len = -1);
  int GetCol(int col, float *d, int base = 0, int len = -1);
  int SetCol(int col, float *d, int base = 0, int len = -1);
  floatArray At(int baser = 0, int basec = 0, int rows = -1, int cols = -1);
	// construct a new array as the specified submatrix
  int Sum(float *rslt);   // sum all columns into float vector
  void Print(char *title = 0);
  void Resize(int rows, int cols);    // change size but retain current data
  void Clear(void);
  void Transpose(void);
  int loadDAT(const char *filename, int dataset=0);	// returns 0 on fail
  int saveDAT(const char *filename, int dataset=0, int trunc=1);	
                           // returns 0 on fail.  <trunc> forces rewrite
  void NormRows(void);	// normalize each row to unit maximum
  void NormCols(void);	// normalize each col to unit maximum
  void Norm(void);	// normalize whole array to unit maximum
    float Min(void);	// overall minimum
    float Max(void);	// overall max
    void  Scale(double s);	// scale whole
    void  Offset(double o);	// shift whole
};

#endif /* _FLOATARRAY_H */
