#!./sowish -file
#
# tkdat
#
# view a dat file using tk
#
# dpwe@media.mit.edu 1995jan30
#

if { $argc < 2 } {
    puts "usage: tkdat <soundfile> <datfile>"
    exit
}

# Make sure libtksound is loaded if this is plain wish
if {[info commands sound] == ""} {
    set libfile $env(HOME)/lib/$env(SPEECH_ARCH)/libtksound[info sharedlibext]
#    set libfile sun4-sunos5-CC/libtksound.so

    if {[file readable $libfile] && [info commands load] != ""} {
	load $libfile
    } else {
	puts stderr "$argv0: cannot load library $libfile (Tcl version is $tcl_version)"
	exit -1
    }
}

# allow terminal inputs as well
# set tcl_interactive 1

# make sure the window is resizable
wm minsize . 1 1

# load the sound and acg files
set soundfilename [lindex $argv 0]
set datfilename [lindex $argv 1]
set sound so
sound $sound
$sound open $soundfilename
set soDuration [expr "[$sound config -nframes]/[$sound config -samplerate]"]

# Yes, a display
# xpad is supposed to be size of scrollbar arrows
set xpad 13
# 16 if scrollbar relief is sunken

proc configval {widget param} {
    # it's tiresome and clumsy to extract the current value from the 
    # standard 'config' response - so do it with this function instead
    return [lindex [$widget config $param] 4]
}

proc scrollbarCommand {baseval} {
    global dsets datfilename
    if { $baseval < 0 }  {
	set baseval 0
    } elseif { $baseval >= $dsets } {
	set baseval [expr "$dsets -1"]
    }
    fa load $datfilename $baseval
    global normCols
    # $normCols is set by button
    if {$normCols} {
	fa normcols
    }
    .im rescan
#    .im config -dset $baseval
    .sb set $dsets 1 $baseval $baseval
    .buttons.zoomsel config -text $baseval
}

proc sndwvfmSelLims {sndview {noneisall 0}} {
    # return the selection base and len.
    # If $noneisall is present and nonzero, return the entire sound limits 
    # if the selection is zero len
    set selBase  [configval $sndview -selBase]
    set selLen   [configval $sndview -selLen ]
    set viewBase [configval $sndview -viewBase]
    set viewLen  [configval $sndview -viewLen ]
    set sound    [configval $sndview -sound]
    set maxLen   [$sound config -nframes]
    if { $noneisall && ($selLen == 0) } {
	set selBase 0
	set selLen $maxLen
    }
    return [list $selBase $selLen $viewBase $viewLen $maxLen]
}

proc svplay {sndview} {
    # play the selected region of the soundview
    set sound   [configval $sndview -sound]
    set sellims [sndwvfmSelLims $sndview 1]
    set selBase [lindex $sellims 0]
    set selLen  [lindex $sellims 1]
    $sound play $selBase $selLen
}

proc sliderproc {root subcommand val} {
    # command for default slider
    $root.value config -text $val
    # have to eval to cause reparsing of $subcommand into separate elements
    eval "$subcommand $val"
}

proc defslider {root label command {initval "0"}} {
    # Define a slider in a frame with a label
    frame $root -borderwidth 2 -relief groove
    scale $root.sc -to 0 -from 99 -showvalue 0 \
	    -command [list sliderproc $root $command]
    label $root.label -text $label
    label $root.value -text $initval
    pack $root.label -side top    -fill both -padx 5 -pady 5
    pack $root.value -side bottom -fill both -padx 5 -pady 5
    pack $root.sc    -side top    -fill both -expand y -padx 5 -pady 5
    $root.sc set $initval
    return $root
}

proc setBriScale {img val} {
    # scale callback for brightness widget - map 1..99 to -1..1
    $img config -offset [expr "$val/50.0-1.0"]
    $img rescan
}

proc setConScale {img val} {
    # scale callback for contrast widget - map 0..99 to 0.1x .. 10x
    
    # precondition val so that 50 gives 1.0 x
    set val   [expr "int($val - 50)"]

    set range [expr "$val/6"]
    set pten  [expr "pow(10.0,$range)"]
    set arg   [lindex {1.0 1.5 2.0 3.0 5.0 7.0} [expr "$val - 6*$range"]]
    set scale [expr "$arg * $pten"]

    $img config -scale $scale
    $img rescan
}

# float array is connected to DAT
farray fa
fa load $datfilename

##########  Initialize all the widgets #############
sndwvfm   .ov     -sound $sound -padX $xpad -padY 5 -height 100 \
		  -viewLen [$sound config -nframes]
timeline  .ovtl   -padX $xpad -padY 0 -height 20 -maxVal $soDuration \
		  -tixTopLeft 0 -relief flat
image     .im      -transpose 1 -farray fa -padX $xpad -padY 5 -relief flat
label     .status -text "File: $datfilename"
scrollbar .sb     -orient horiz -relief flat -command "scrollbarCommand"

frame .cf
defslider .cf.brisli "bri" "setBriScale .im" 50
defslider .cf.consli "co" "setConScale .im" 50
set normCols 1
checkbutton .cf.norm -text "norm" -variable normCols
pack .cf.norm -side bottom -fill x -expand 0
pack .cf.brisli .cf.consli -side left -fill both -expand 1

set $sound.views ".ov"

frame  .buttons
button .buttons.zoomin  -text " < "  
#-command "zoom .sw .7"
button .buttons.zoomout -text " > "  
#-command "zoom .sw 1.4"
button .buttons.zoomsel -text "zoom" 
#-command "zoomsel .sw"
button .buttons.quit    -text "quit" -command "exit"
button .buttons.play    -text "play" -command "svplay .ov"
#-command "svplay .sw"

######### Pack all the widgets ###########
pack .status          -side top -fill x -expand no
pack .ovtl  -side top    -fill x
pack .ov    -side top    -fill x
pack .sb    -side top    -fill x -expand no 

pack .buttons         -side bottom -fill x -expand no -padx 6 -pady 6
pack .buttons.zoomin  -side left
pack .buttons.zoomout -side right
pack .buttons.play .buttons.zoomsel .buttons.quit -side left -expand yes

#pack .brisli          -side left   -fill y -expand no
#pack .consli          -side left   -fill y -expand no
pack .cf              -side left   -fill y    -expand no

pack .im              -side bottom -fill both -expand yes

set finfo [datinfo $datfilename]
set dsets [ lindex $finfo 0]
set rs    [ lindex $finfo 2]
set cs    [ lindex $finfo 3]
.im config -rlen $rs -clen $cs

######## Selection and autoscrolling functions ############

proc scroll {sndview percent} {
    # shift the current view by some percentage of the current view - 
    #   forward if $percent > 0, else backward
    set viewBase [configval $sndview -viewBase]
    set viewLen  [configval $sndview -viewLen ]
    set newBase [expr "$viewBase + ($viewLen * $percent)/100"]
    $sndview config -viewBase $newBase
}

proc startSel {sndview xpix}  {
    # mouse down in view
    global anchor$sndview
    set frame [$sndview frameOfPixel $xpix]
    # set new selection in all concerned views
    set viewsname [configval $sndview -sound].views
    global $viewsname
    foreach view [set $viewsname] {
       $view config -selLen 0 -selBase $frame
    }
    set anchor$sndview $frame
}

# semaphore to pre-empt auto-repeats
set sndwvfm_auto_pending 0
set sndwvfm_cancel_auto  0

proc endSel {sndview} {
    # called when the scroll ends: if an auto pending, be sure to cancel it
    global sndwvfm_auto_pending sndwvfm_cancel_auto
    if { $sndwvfm_auto_pending }  {
	set sndwvfm_cancel_auto 1
    }
}

proc scrollGrowSel {sndview scrollamt}  {
    # scroll the view and act as if the button was moved, since it 
    # has moved, relative to the scrolled view.
    # This command is executed by a timer a little while after the 
    # selection is grown into the 'end zones'
    global sndwvfm_auto_pending sndwvfm_cancel_auto currentX$sndview
    set sndwvfm_auto_pending 0
    if { $sndwvfm_cancel_auto } {
	# this autoscroll event was cancelled before it ever got to happen
	set sndwvfm_cancel_auto 0
    } else {
	scroll $sndview $scrollamt
	# re-execute the scroll .. 
	growSel $sndview [set currentX$sndview]
    }
}

proc checkForAuto {sndview xpix}  {
    # if we're near the edge of the view, schedule an auto-scroll
    # ('near' is within $border pixels)
    global sndwvfm_auto_pending sndwvfm_cancel_auto
    set border 50
    set lowerlim $border
    set upperlim [expr "[winfo width $sndview] - $border"]
    if { $xpix < $lowerlim || $xpix > $upperlim } {
	# calculate the scroll amount so that it depends how far the 
	# mouse is into the autoscroll region
	if { $xpix < $lowerlim } {
	    set xoffset [expr "$xpix - $lowerlim"]
	} else {
	    set xoffset [expr "$xpix - $upperlim"]
	}
	set scrollamt [expr "(5 * $xoffset)/$border"]
	if { $sndwvfm_auto_pending == 0 } {
	    after 200 "scrollGrowSel $sndview $scrollamt"
	    set sndwvfm_auto_pending 1
	}
	# just in case we have cancelled a pending auto, re-instance it
	set sndwvfm_cancel_auto 0
    } else {
	# we have moved out of the autoscroll zone; cancel any pending scroll
	if { $sndwvfm_auto_pending } {
	    set sndwvfm_cancel_auto 1
	}
    }
}

proc growSel {sndview xpix}  {
    # mouse drag in view - just extend the selection
    global anchor$sndview currentX$sndview
    # save the current x position for the redraw to figure
    set currentX$sndview $xpix
    # better set up for an autoscroll if we need one
    checkForAuto $sndview $xpix
    # OK, now just grow the selection
    set origPos [set anchor$sndview]
    set newPos  [$sndview frameOfPixel $xpix]
    if { $newPos < $origPos } {
	set tmp $newPos
	set newPos $origPos
	set origPos $tmp
    }
    set newLen [expr "$newPos - $origPos"]
    # set new selection in all involved windows
    set viewsname [configval $sndview -sound].views
    global $viewsname
    foreach view [set $viewsname] {
       $view config -selBase $origPos -selLen $newLen
    }
}

proc startSelExtend {sndview xpix}  {
    # mouse down in view, but keep more distant of current anchor points
    global anchor$sndview
    set selBase [configval $sndview -selBase]
    set selLen  [configval $sndview -selLen ]
    set selEnd  [expr "$selBase + $selLen"]
    set newPos  [$sndview frameOfPixel $xpix]
    if {$selLen == 0}  {
	# no selection to extend - act like button 1
	startSel $sndview $xpix
    } else {
	# set the anchor to the more distant of the selection points
        if { abs($newPos - $selBase) > abs($newPos - $selEnd) } {
	    set anchor$sndview $selBase
	} else {
	    set anchor$sndview $selEnd
	}
	# and now act as if cursor just moved
	growSel $sndview $xpix
    }
}

bind .ov <1> 			{startSel .ov %x}
bind .ov <B1-Motion> 		{growSel .ov %x}
bind .ov <B1-ButtonRelease> 	{endSel .ov}
bind .ov <3> 			{startSelExtend .ov %x}
bind .ov <B3-Motion> 		{growSel .ov %x}
bind .ov <B3-ButtonRelease> 	{endSel .ov}
bind .ov <2> 			{zoomsel .sw}

scrollbarCommand 0
