#!./sowish -file
# test script for sound functions

# Make sure libtksound is loaded if this is plain wish
if {[info commands sound] == ""} {
    set libfile $env(HOME)/lib/$env(SPEECH_ARCH)/libtksound[info sharedlibext]
    if {[file readable $libfile] && [info commands load] != ""} {
	load $libfile
    } else {
	puts stderr "$argv0: cannot load library $libfile (Tcl version is $tcl_version)"
	exit -1
    }
}

#set filename "sh2m.aif"
set filename [lindex $argv 0]

sound so
so open $filename
# so play
set sound so
set soDuration [expr "[so config -nframes]/[so config -samplerate]"]

# make sure the window is resizable
wm minsize . 1 1

proc timelineAndScrollCommand {timeline scrollbar sound totConts viewSize top bottom} {
    # normally the soundview scrollback command plugs straight into 
    # the scrollbar - but here we steal off a bit to feed the time line 
    # as well
    set srate [$sound config -samplerate]
    set tmin  [expr "$top / $srate"]
    set tmax  [expr "($bottom + 1)/$srate"]
#    $timeline config -minVal $tmin -maxVal $tmax
    $scrollbar set $totConts $viewSize $top $bottom
}

# xpad is supposed to be size of scrollbar arrows
set xpad 19

label .status -text "File: $filename"
scrollbar .sb -orient horiz -relief flat -command ".sw config -viewBase "
sndwvfm .ov -sound $sound -padX $xpad -padY 5 -height 100 -viewLen [$sound config -nframes]
timeline .ovtl -padX $xpad -padY 0 -height 20 -maxVal $soDuration -tixTopLeft 0 -relief flat
#timeline .swtl -padX $xpad -padY 0 -height 20 -maxVal $soDuration -tixTopLeft 1 -relief flat
sndwvfm .sw -sound $sound -padX $xpad -padY 5 -viewLen 10000 -scrollCommand {timelineAndScrollCommand .swtl .sb [lindex [.sw config -sound] 4]}
# ".sb set"

# global list of views associated with this sound
set $sound.views ".sw .ov"

frame .buttons
button .buttons.zoomin -text " < " -command "zoom .sw .7"
button .buttons.zoomout -text " > " -command "zoom .sw 1.4"
button .buttons.zoomsel -text "zoom" -command "zoomsel .sw"
button .buttons.quit  -text "quit" -command "exit"
button .buttons.play  -text "play" -command "svplay .sw"

pack .status -side top -fill x -expand no
pack .buttons  -side bottom -fill x -expand no -padx 6 -pady 6
pack .buttons.zoomin -side left
pack .buttons.zoomout -side right
pack .buttons.play .buttons.zoomsel .buttons.quit -side left -expand yes

pack .ovtl -side top -fill x
pack .ov -side top -fill x
pack .sb -side top -fill x
#pack .swtl -side bottom -fill x
pack .sw -side bottom -expand yes -fill both

proc scroll {sndview percent} {
    # shift the current view by some percentage of the current view - 
    #   forward if $percent > 0, else backward
    set viewBase [lindex [$sndview config -viewBase] 4]
    set viewLen  [lindex [$sndview config -viewLen ] 4]
    set newBase [expr "$viewBase + ($viewLen * $percent)/100"]
    $sndview config -viewBase $newBase
}

proc startSel {sndview xpix}  {
    # mouse down in view
    global anchor$sndview
    set frame [$sndview frameOfPixel $xpix]
    # set new selection in all concerned views
    set viewsname [lindex [$sndview config -sound] 4].views
    global $viewsname
    foreach view [set $viewsname] {
       $view config -selLen 0 -selBase $frame
    }
    set anchor$sndview $frame
}

# semaphore to pre-empt auto-repeats
set sndwvfm_auto_pending 0
set sndwvfm_cancel_auto  0

proc endSel {sndview} {
    # called when the scroll ends: if an auto pending, be sure to cancel it
    global sndwvfm_auto_pending sndwvfm_cancel_auto
    if { $sndwvfm_auto_pending }  {
	set sndwvfm_cancel_auto 1
    }
}

proc scrollGrowSel {sndview scrollamt}  {
    # scroll the window and act as if the button was pressed.
    # This command is executed by a timer a little while after the 
    # selection is grown into the 'end zones'
    global sndwvfm_auto_pending sndwvfm_cancel_auto sndwvfm_last_x
    set sndwvfm_auto_pending 0
    if { $sndwvfm_cancel_auto } {
	set sndwvfm_cancel_auto 0
    } else {
	scroll $sndview $scrollamt
	# re-execute the scroll .. 
	growSel $sndview $sndwvfm_last_x
    }
}

proc growSel {sndview xpix {autoscroll 0}}  {
    # mouse drag in view - extend the selection
    # auto is a flag to indicate that this call is the result of 
    # an autorepeat
    global anchor$sndview 
    global sndwvfm_auto_pending sndwvfm_cancel_auto sndwvfm_last_x
    # save the value of this instance for re-use by auto-repeat
    set sndwvfm_last_x $xpix
    # if we're near the edge of the view, setup auto-scroll
    # ('near' is within $border pixels)
    set border 50
    set lowerlim $border
    set upperlim [expr "[winfo width $sndview] - $border"]
    if { $xpix < $lowerlim || $xpix > $upperlim } {
	set scrollamt [expr "($xpix<$lowerlim)?-5:5"]
	if { $sndwvfm_auto_pending == 0 } {
	    after 200 "scrollGrowSel $sndview $scrollamt"
	    set sndwvfm_auto_pending 1
	}
	# just in case we have cancelled a pending auto, re-instance it
	set sndwvfm_cancel_auto 0
    } else {
	# we have moved out of the danger zone, so cancel any pending scroll
	if { $sndwvfm_auto_pending } {
	    set sndwvfm_cancel_auto 1
	}
    }
    # OK, now just grow the selection
    set origPos [set anchor$sndview]
    set newPos  [$sndview frameOfPixel $xpix]
    if { $newPos < $origPos } {
	set tmp $newPos
	set newPos $origPos
	set origPos $tmp
    }
    set newLen [expr "$newPos - $origPos"]
    # set new selection in all involved windows
    set viewsname [lindex [$sndview config -sound] 4].views
    global $viewsname
    foreach view [set $viewsname] {
       $view config -selBase $origPos -selLen $newLen
    }
}

proc startSelExtend {sndview xpix}  {
    # mouse down in view, but keep more distant of current anchor points
    global anchor$sndview
    set selBase [lindex [$sndview config -selBase] 4]
    set selLen  [lindex [$sndview config -selLen ] 4]
    set selEnd  [expr "$selBase + $selLen"]
    set newPos  [$sndview frameOfPixel $xpix]
    if {$selLen == 0}  {
	# no selection to extend - act like button 1
	startSel $sndview $xpix
    } else {
	# set the anchor to the more distant of the selection points
        if { abs($newPos - $selBase) > abs($newPos - $selEnd) } {
	    set anchor$sndview $selBase
	} else {
	    set anchor$sndview $selEnd
	}
	# and now act as if cursor just moved
	growSel $sndview $xpix
    }
}

bind .sw <1> {startSel .sw %x}
bind .sw <B1-Motion> {growSel .sw %x}
bind .sw <B1-ButtonRelease> {endSel .sw}
bind .sw <3> {startSelExtend .sw %x}
bind .sw <B3-Motion> {growSel .sw %x}
bind .sw <B3-ButtonRelease> {endSel .sw}
# dspB binding - middle button zooms to mark
bind .sw <2> {zoomsel .sw}

bind .ov <1> {startSel .ov %x}
bind .ov <B1-Motion> {growSel .ov %x}
bind .ov <B1-ButtonRelease> {endSel .ov}
bind .ov <3> {startSelExtend .ov %x}
bind .ov <B3-Motion> {growSel .ov %x}
bind .ov <B3-ButtonRelease> {endSel .ov}
# dspB binding - middle button zooms to mark
bind .ov <2> {zoomsel .sw}

proc sndwvfmSelLims {sndview {noneisall 0}} {
    # return the selection base and len.
    # If $noneisall is present and nonzero, return the entire sound limits 
    # if the selection is zero len
    set selBase [lindex [$sndview config -selBase] 4]
    set selLen  [lindex [$sndview config -selLen ] 4]
    set viewBase [lindex [$sndview config -viewBase] 4]
    set viewLen  [lindex [$sndview config -viewLen ] 4]
    set sound [lindex [$sndview config -sound] 4]
    set maxLen [$sound config -nframes]
    if { $noneisall && ($selLen == 0) } {
	set selBase 0
	set selLen $maxLen
    }
    return [list $selBase $selLen $viewBase $viewLen $maxLen]
}

proc zoom {sndview factor} {
    # adjust the display length of a sndview by the specified factor
    # if factor > 1.0, zooming out
    # attempt to keep view centered at same point
    set sellims [sndwvfmSelLims $sndview 1]
    set viewBase [lindex $sellims 2]
    set viewLen  [lindex $sellims 3]
    set newLen   [expr "int($factor * $viewLen)"]
    set newBase  [expr "$viewBase + ($viewLen - $newLen)/2"]
    # newBase may go negative, but $sndview config will clip it nicely
    $sndview config -viewLen $newLen -viewBase $newBase
}

proc zoomsel {sndview} {
    # zoom sound view to current selection - or whole sound if none
    set sellims [sndwvfmSelLims $sndview 1]
    set selBase [lindex $sellims 0]
    set selLen  [lindex $sellims 1]
    set viewBase [lindex $sellims 2]
    set viewLen  [lindex $sellims 3]
    set maxLen   [lindex $sellims 4]
    # if we're already zoomed to sel, zoom out to full (toggle)
    if { ($viewBase == $selBase) && ($viewLen == $selLen) } {
	set selBase 0
	set selLen $maxLen
    }
    $sndview config -viewBase $selBase -viewLen $selLen
}

proc svplay {sndview} {
    # play the selected region of the soundview
    set sound [lindex [$sndview config -sound] 4]
    set sellims [sndwvfmSelLims $sndview 1]
    set selBase [lindex $sellims 0]
    set selLen  [lindex $sellims 1]
    $sound play $selBase $selLen
}

set tcl_interactive 1

